import datetime
import sys

from .exceptions import AdmRawError, NotSavedError
from .item import ClassItemABC
from .score import ExerciseScore, SheetScore
from .user import User
from .utils import default, one_year_later



class Sheet(ClassItemABC):
    """This class is used to represent a WIMS' Worksheet.
    
    Parameters:
        title - (str) name of the sheet (defaults to "sheet n#")
        description - (str) description of the sheet (defaults to "sheet n#")
        expiration - (str) expiration date (yyyymmdd) defaults to one year later
        sheetmode - (str) the mode of the sheet:
            0 : pending (default)
            1 : active
            2 : expired
            3 : expired + hidden
        weight - (int) weight of the sheet in the class score (default to 1), use 0 if you want this
            sheet's score to be ignored.
        formula - (str) How the score is calculated for this sheet (0 to 6, default to 2)
            0 : Very lenient: maximum between percentage and quality.
            1 : Quality is not taken into account. You get maximum of grade once all
                the required work is done.
            2 : Quality has only slight effect over the grade.
            3 : More effect of quality.
            4 : To have a grade of 10, you must gather all required points (100%)
                without making any error (quality=10).
            5 : Unfinished work is over-punished.
            6 : Any mistake is over-punished.
        indicator - (str) what indicator will be used in the score formula (0 to 2, default to 1)
        contents - (str) the contents for the multi-line file to be created.
            The semicolons (;) in this parameter will be interpreted as new lines. Equal characters
            (=) must be replaced by the character AT (@). There is no check made, so the integrity
            of the contents is up to you only! (defaults to "")"""
    
    
    def __init__(self, title=None, description=None, expiration=None, sheetmode=0, weight=1,
                 formula=2, indicator=1, contents="", **kwargs):
        if expiration is not None:
            datetime.datetime.strptime(expiration, "%Y%m%d")
        
        super().__init__()
        self._class = None
        self._saved = False
        self.wclass = False
        self.qsheet = sys.maxsize
        self.title = title
        self.description = description
        self.expiration = expiration if expiration is not None else one_year_later()
        self.sheetmode = sheetmode
        self.weight = weight
        self.formula = formula
        self.indicator = indicator
        self.contents = contents
        self.exos = []
    
    
    @property
    def exo_count(self):
        """Returns the number of exercises in this sheet."""
        return len(self.exos)
    
    
    @property
    def infos(self):
        """Return all the informations hosted on the WIMS server about this sheet."""
        if not self._class:
            raise NotSavedError("infos is not defined until the sheet has been saved once")
        
        status, sheet_info = self._class._api.getsheet(
            self._class.qclass, self._class.rclass, self.qsheet, verbose=True)
        if not status:
            raise AdmRawError(sheet_info['message'])
        
        for k in ['status', 'code', 'job']:
            del sheet_info[k]
        
        return sheet_info
    
    
    def __str__(self):
        return "<wimsapi.Sheet object at %s - qsheet : %s>" % (hex(id(self)), str(self.qsheet))
    
    
    __repr__ = __str__
    
    
    def __eq__(self, other):
        """Sheets have to come from the same server and have the same qsheet to be equal."""
        if isinstance(other, self.__class__):
            if not self.wclass or not other.wclass:
                raise NotSavedError("Cannot test equality between unsaved sheets")
            return str(self.qsheet) == str(other.qsheet) and self._class == other._class
        return False
    
    
    def __hash__(self):
        if not self.wclass:
            raise NotSavedError("Unsaved User cannot be hashed")
        return hash((self._class.qclass, self.qsheet))
    
    
    def refresh(self):
        """Refresh this instance of a WIMS Sheet from the server itself."""
        if not self.wclass:
            raise NotSavedError("Can't refresh unsaved sheet")
        
        new = Sheet.get(self._class, self.qsheet)
        self.__class__ = new.__class__
        self.__dict__ = new.__dict__
        
        return self
    
    
    def _to_payload(self):
        return {k: v for k, v in self.__dict__.items() if k not in ['qsheet', '_saved', '_class']}
    
    
    def save(self, wclass=None, check_exists=True):
        """Save the Sheet in the given class.

        wclass is an instance of wimsapi.Class. The argument is optionnal
        if the sheet has already been saved or fetched from a Class, it
        will default to the last Class used to saved or the Class from which
        the sheet was fetched.
        
        If check_exists is True, the api will check if a sheet with the same ID
        exists on the WIMS' server. If it exists, save will instead modify this
        sheet instead of trying to create new one."""
        if not wclass and not self._class:
            raise NotSavedError("wclass must be provided if this sheet has neither been imported "
                                "from a WIMS class nor saved once yet")
        
        wclass = wclass or self._class
        
        if not wclass._saved:
            raise NotSavedError("Class must be saved before being able to save a sheet")
        
        if wclass is not None and check_exists:
            self._saved = wclass.checkitem(self)
        
        if self._saved:
            status, response = wclass._api.modsheet(
                wclass.qclass, wclass.rclass, self.qsheet, self._to_payload(), verbose=True)
        else:
            status, response = wclass._api.addsheet(
                wclass.qclass, wclass.rclass, self._to_payload(), verbose=True)
        
        if not status:
            raise AdmRawError(response['message'])
        
        self.qsheet = response['sheet_id'] if "sheet_id" in response else response["querysheet"]
        self._class = wclass
        self.wclass = True
    
    
    def delete(self):
        """Delete the sheet from its associated WIMS class on the server."""
        if not self.wclass:
            raise NotSavedError("Cannot delete an unsaved sheet")
        
        status, response = self._class._api.delsheet(self._class.qclass, self._class.rclass,
                                                     self.qsheet, verbose=True)
        if not status:
            raise AdmRawError(response['message'])
        
        self.wclass = False
        self._class = None
    
    
    @classmethod
    def check(cls, wclass, sheet):
        """Returns True if sheet is in wclass, False otherwise.

        sheet can be either an instance of Sheet, or an int corresponding to the identifier (qsheet)
        of the Sheet in the WIMS class.

        E.G. either Sheet.check(wclass, 1) or Sheet.check(wclass, Sheet(...))"""
        if not wclass._saved:
            raise NotSavedError("Class must be saved before being able to check whether a sheet "
                                "exists")
        
        sheet = sheet.qsheet if isinstance(sheet, cls) else sheet
        status, response = wclass._api.checksheet(wclass.qclass, wclass.rclass, sheet,
                                                  verbose=True)
        msg = ('element #%s of type sheet does not exist in this class (%s)'
               % (str(sheet), str(wclass.qclass)))
        if not status and msg not in response['message']:  # pragma: no cover
            raise AdmRawError(response['message'])
        
        return status
    
    
    @classmethod
    def remove(cls, wclass, sheet):
        """Remove the sheet from wclass.

        sheet can be either an instance of Sheet, or an int corresponding to the identifier (qsheet)
        of the Sheet in the WIMS class.

        E.G. either Sheet.remove(wclass, 1) or Sheet.remove(wclass, Sheet(...))"""
        if not wclass._saved:
            raise NotSavedError("Class must be saved before being able to remove a sheet")
        
        qsheet = sheet.qsheet if isinstance(sheet, cls) else sheet
        status, response = wclass._api.delsheet(wclass.qclass, wclass.rclass, qsheet, verbose=True)
        if not status:
            raise AdmRawError(response['message'])
    
    
    @classmethod
    def get(cls, wclass, qsheet):
        """Returns an instance of Sheet corresponding to qsheet in wclass."""
        if not wclass._saved:
            raise NotSavedError("Class must be saved before being able to get a sheet")
        
        status, sheet_info = wclass._api.getsheet(wclass.qclass, wclass.rclass, qsheet,
                                                  verbose=True)
        if not status:
            raise AdmRawError(sheet_info['message'])
        
        duplicate = dict(sheet_info)
        for k, v in duplicate.items():
            if k.startswith("sheet_"):
                sheet_info[k[6:]] = v
        sheet_info["sheetmode"] = sheet_info["sheet_status"]
        
        sheet = Sheet(**sheet_info)
        sheet.qsheet = sheet_info["query_sheet"]
        sheet._class = wclass
        sheet.wclass = True
        return sheet
    
    
    @classmethod
    def list(cls, wclass):
        """Returns a list of every Sheet of wclass."""
        status, response = wclass._api.listsheets(wclass.qclass, wclass.rclass, verbose=True)
        if not status:
            raise AdmRawError(response['message'])
        
        return [cls.get(wclass, qsheet) for qsheet in response["sheetlist"] if qsheet != '']
    
    
    @staticmethod
    def _compute_grade(formula, i, Q, cumul, best, acquired):
        """Compute the grade of a sheet according to the formula and the chosen I.
        
        Formula contains both Q and I."""
        i = int(i)
        if i == 0:
            I = cumul
        elif i == 1:
            I = best
        else:
            I = acquired
        
        Q /= 10
        I /= 100
        formula = "10 * (%s)" % formula
        return round(eval(formula.replace("^", "**")), 2)
    
    
    def scores(self, user=None):
        """Returns a list of SheetScore for every user. If user is given returns only its
        SheetScore.
        
        user can either be an instance of wimsapi.User or its quser.
        
        A value of -1 on some members mean that WIMS did not send the value. This may be caused
        by an outdated WIMS server."""
        if not self.wclass:
            raise NotSavedError("Sheet must be saved before being able to retrieve scores")
        
        quser = user.quser if isinstance(user, User) else user
        if quser is not None and not self._class.checkitem(quser, User):  # Checks that quser exists
            raise ValueError("User '%s' does not exists in class '%s'" % (user, self._class.qclass))
        
        status, response = self._class._api.getsheetscores(self._class.qclass, self._class.rclass,
                                                           self.qsheet, verbose=True)
        if not status:
            raise AdmRawError(response['message'])
        
        scores = []
        exo_count = len(response["exo_weights" if "exo_weights" in response else "weights"])
        for data in response["data_scores"]:
            if quser is not None and data['id'] != quser:  # Searching for specific user score
                continue
            
            user = self._class.getitem(data['id'], User)
            try:
                score = self._compute_grade(response["sheet_formula"]["formula"],
                                            response["sheet_formula"]["I"], data["user_quality"],
                                            data["user_percent"], data["user_best"],
                                            data["user_level"])
            except Exception:  # pragma: no cover
                score = -1
            
            sheet_params = {
                "sheet":     self,
                "user":      user,
                "score":     score,
                "quality":   data.get("user_quality", -1),
                "cumul":     data.get("user_percent", -1),
                "best":      data.get("user_best", -1),
                "acquired":  data.get("user_level", -1),
                "weight":    self.weight,
                "exercises": [],
            }
            for i in range(exo_count):
                exo_params = {
                    "exo":      None,
                    "user":     user,
                    "quality":  default(data, "mean_detail", i, -1),
                    "cumul":    default(data, "got_detail", i, -1),
                    "best":     default(data, "best_detail", i, -1),
                    "acquired": default(data, "level_detail", i, -1),
                    "last":     default(data, "last_detail", i, -1),
                    "tries":    default(data, "try_detail", i, -1),
                    "weight":   default(response, "exo_weights", i, -1),
                    "required": default(response, "requires", i, -1),
                }
                sheet_params["exercises"].append(ExerciseScore(**exo_params))
            
            scores.append(SheetScore(**sheet_params))
        
        return scores[0] if quser is not None else scores



import datetime
import os
import unittest
from unittest import mock

from wimsapi.api import WimsAPI
from wimsapi.exceptions import AdmRawError, InvalidItemTypeError, NotSavedError
from wimsapi.sheet import Sheet
from wimsapi.user import User
from wimsapi.wclass import Class, one_year_later


WIMS_URL = os.getenv("WIMS_URL") or "http://localhost:7777/wims/wims.cgi/"



class FakeDate(datetime.date):
    """Used to override datetime.date.today"""
    
    
    @classmethod
    def today(cls):
        """Always return a Date corresponding to 1966-11-16."""
        return datetime.date(1966, 11, 16)



class ClassTestCase(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        """Create an API and an User to use through the tests."""
        cls.api = WimsAPI(WIMS_URL, "myself", "toto")
        cls.user = User("supervisor", "last", "first", "pass", "mail@mail.com")
    
    
    def tearDown(self):
        """Delete classes that a test might have created."""
        self.api.delclass(999999, "myclass")
        self.api.delclass(999666, "myclass")
        self.api.delclass(999990, "myclass")
    
    
    @mock.patch('datetime.date', FakeDate)
    def test_one_year_later(self):
        self.assertEqual(one_year_later(), "19671116")
    
    
    def test_init_and_properties(self):
        c = Class.get(WIMS_URL, "myself", "toto", 9001, "myclass")
        self.assertEqual(c.url, WIMS_URL)
        self.assertEqual(c.ident, "myself")
        self.assertEqual(c.passwd, "toto")
        self.assertIn("description", c.infos)
        
        c = Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999)
        with self.assertRaises(NotSavedError):
            c.url
        with self.assertRaises(NotSavedError):
            c.ident
        with self.assertRaises(NotSavedError):
            c.passwd
        with self.assertRaises(NotSavedError):
            c.infos
        
        with self.assertRaises(ValueError):
            Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999, lang="Wrong")
        with self.assertRaises(ValueError):
            Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999, level="Wrong")
        with self.assertRaises(ValueError):
            Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999, expiration="Wrong")
    
    
    def test_append_slash(self):
        url = WIMS_URL if not WIMS_URL.endswith('/') else WIMS_URL[:-1]
        c = Class.get(url, "myself", "toto", 9001, "myclass")
        self.assertEqual(c.url, url + "/")
    
    
    def test_save_and_refresh(self):
        c = Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999)
        
        with self.assertRaises(NotSavedError):
            c.save()
        
        with self.assertRaises(NotSavedError):
            c.refresh()
        
        c.save(WIMS_URL, "myself", "toto")
        c = Class.get(WIMS_URL, "myself", "toto", 999999, "myclass")
        c2 = Class.get(WIMS_URL, "myself", "toto", 999999, "myclass")
        
        self.assertEqual(c.institution, "an institution")
        self.assertEqual(c2.institution, "an institution")
        
        c.institution = "modified"
        c.save()
        self.assertEqual(c.institution, "modified")
        self.assertEqual(c2.institution, "an institution")
        
        c2.refresh()
        self.assertEqual(c2.institution, "modified")
        self.api.delclass(99999999, "myclass")
    
    
    def test_save_without_qclass(self):
        c = Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user)
        
        with self.assertRaises(NotSavedError):
            c.save()
        
        c.save(WIMS_URL, "myself", "toto")
        self.assertIsNotNone(c.qclass)
        c.delete()
    
    
    def test_getitem(self):
        c = Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999)
        
        with self.assertRaises(NotSavedError):
            c.getitem("supervisor", User)
        with self.assertRaises(InvalidItemTypeError):
            c.getitem(1, int)
        
        c.save(WIMS_URL, "myself", "toto")
        user = c.getitem("supervisor", User)
        self.assertEqual(user.firstname, self.user.firstname)
    
    
    def test_checkitem(self):
        c = Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999)
        unknown = User("unknown", "last", "first", "pass", "mail2@mail.com")
        
        with self.assertRaises(NotSavedError):
            c.checkitem("supervisor", User)
        with self.assertRaises(InvalidItemTypeError):
            c.checkitem(1)
        
        c.save(WIMS_URL, "myself", "toto")
        self.assertTrue(c.checkitem("supervisor", User))
        self.assertTrue(c.checkitem(self.user))
        self.assertFalse(c.checkitem("Unknown", User))
        self.assertFalse(c.checkitem(unknown))
    
    
    def test___contains__(self):
        c = Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999)
        unknown = User("unknown", "last", "first", "pass", "mail2@mail.com")
        
        with self.assertRaises(NotSavedError):
            unknown in c
        
        c.save(WIMS_URL, "myself", "toto")
        self.assertTrue(self.user in c)
        self.assertFalse(unknown in c)
    
    
    def test_additem(self):
        c = Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999)
        u = User("quser", "last", "first", "pass", "mail2@mail.com")
        
        with self.assertRaises(NotSavedError):
            c.additem(u)
        with self.assertRaises(InvalidItemTypeError):
            c.additem(int)
        
        c.save(WIMS_URL, "myself", "toto")
        c.additem(u)
        self.assertEqual(u._class.qclass, c.qclass)
        self.assertEqual(u.wclass, True)
        
        u2 = c.getitem("quser", User)
        self.assertEqual(u2.firstname, u.firstname)
    
    
    def test_delitem(self):
        c = Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999)
        u = User("quser", "last", "first", "pass", "mail2@mail.com")
        u2 = User("quser2", "last", "first", "pass", "mail2@mail.com")
        
        with self.assertRaises(NotSavedError):
            c.delitem(u)
        with self.assertRaises(InvalidItemTypeError):
            c.delitem(int)
        
        c.save(WIMS_URL, "myself", "toto")
        c.additem(u)
        c.additem(u2)
        
        self.assertTrue(c.checkitem("quser", User))
        c.delitem(u)
        self.assertFalse(c.checkitem("quser", User))
        
        self.assertTrue(c.checkitem("quser2", User))
        c.delitem("quser2", User)
        self.assertFalse(c.checkitem("quser2", User))
    
    
    def test_delete(self):
        c = Class("myclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999)
        
        with self.assertRaises(NotSavedError):
            c.delete()
        
        c.save(WIMS_URL, "myself", "toto")
        
        Class.get(WIMS_URL, "myself", "toto", c.qclass, c.rclass)  #  Ok
        c.delete()
        with self.assertRaises(AdmRawError):
            Class.get(WIMS_URL, "myself", "toto", c.qclass, c.rclass)  # Should raise the exception
    
    
    def test_list(self):
        c1 = Class("rclass", "A class", "an institution", "mail@mail.com", "password", self.user)
        c2 = Class("rclass", "A class", "an institution", "mail@mail.com", "password", self.user)
        c3 = Class("rclass", "A class", "an institution", "mail@mail.com", "password", self.user)
        
        c1.save(WIMS_URL, "myself", "toto")
        c2.save(WIMS_URL, "myself", "toto")
        c3.save(WIMS_URL, "myself", "toto")
        
        self.assertListEqual(
            sorted([c1, c2, c3], key=lambda i: i.qclass),
            sorted(Class.list(WIMS_URL, "myself", "toto", "rclass"), key=lambda i: i.qclass)
        )
        
        self.assertListEqual(
            [],
            Class.list(WIMS_URL, "myself", "toto", "unknown_rclass")
        )
        
        c1.delete()
        c2.delete()
        c3.delete()
    
    
    def test_eq(self):
        c1 = Class("rclass", "A class", "an institution", "mail@mail.com", "password", self.user)
        c2 = Class("rclass", "A class", "an institution", "mail@mail.com", "password", self.user)
        c3 = Class("rclass", "A class", "an institution", "mail@mail.com", "password", self.user)
        
        with self.assertRaises(NotSavedError):
            c1 == c3
        
        c1.save(WIMS_URL, "myself", "toto")
        c2.save(WIMS_URL, "myself", "toto")
        c3.save(WIMS_URL, "myself", "toto")
        
        self.assertEqual(c1, Class.get(WIMS_URL, "myself", "toto", c1.qclass, c1.rclass))
        self.assertNotEqual(c2, Class.get(WIMS_URL, "myself", "toto", c1.qclass, c1.rclass))
        self.assertNotEqual(c2, 1)
        
        c1.delete()
        c2.delete()
        c3.delete()
    
    
    def test_listitem(self):
        c = Class("rclass", "A class", "an institution", "mail@mail.com", "password",
                  self.user, qclass=999999)
        
        with self.assertRaises(NotSavedError):
            c.listitem(Sheet)
        with self.assertRaises(InvalidItemTypeError):
            c.listitem(int)
        
        s1 = Sheet("First", "First one")
        s2 = Sheet("Second", "Second one")
        s3 = Sheet("Third", "Third one")
        
        c.save(WIMS_URL, "myself", "toto")
        
        self.assertListEqual([], c.listitem(Sheet))
        
        c.additem(s1)
        c.additem(s2)
        c.additem(s3)
        
        self.assertListEqual(
            sorted([s1, s2, s3], key=lambda i: i.qsheet),
            sorted(c.listitem(Sheet), key=lambda i: i.qsheet))
        
        c.delete()
